/*
 *
 *  Bluetooth low-complexity, subband codec (SBC) library
 *
 *  Copyright (C) 2004  Marcel Holtmann <marcel@holtmann.org>
 *  Copyright (C) 2004  Henryk Ploetz <henryk@ploetzli.ch>
 *
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef __SBC_H
#define __SBC_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

#define SBC_NULL	0x00000001

struct sbc_struct {
	unsigned long flags;

	int rate;
	int channels;
	int blocks;
	int subbands;
	int bitpool;

	void *data;
	int size;
	int len;

	unsigned long duration;

	void *priv;
};

typedef struct sbc_struct sbc_t;

#ifdef USE_FIXED

typedef int32_t sbc_fixed_t;

// this must be an even number
#define FRAC 20

// FRAC/2 leading ones for sign-extended shift
#define PAD (0xffffffff<<(32-(FRAC>>1)))

// FRAC/2 trailing ones for finding the roundoff portion
#define RMASK (0xffffffff>>(32-(FRAC>>1)))

#define FIXED_ONE (1<<FRAC)

// sign-extend half the number of shift (fractional) bits
static inline sbc_fixed_t signextshift(sbc_fixed_t i) { 
	if(i<0) return (i>>(FRAC>>1)) | PAD;
	return i>>(FRAC>>1); 
}

static inline sbc_fixed_t itofixed(int i) { return i<<FRAC; }

static inline int fixedtoi(sbc_fixed_t f) { return f>>FRAC; }

static inline sbc_fixed_t dtofixed(double d) { return (sbc_fixed_t)(d * (double)FIXED_ONE); }

static inline double fixedtod(sbc_fixed_t f) { return f / (double)FIXED_ONE; }

static inline sbc_fixed_t fxmultquick(sbc_fixed_t a, sbc_fixed_t b) { 
	sbc_fixed_t p = signextshift(a)*signextshift(b);
	return p;
}

static inline sbc_fixed_t fxmult(sbc_fixed_t a, sbc_fixed_t b) { 
	// don't lose as much roundoff precision
	sbc_fixed_t ad = signextshift(a);
	sbc_fixed_t bd = signextshift(b);
	sbc_fixed_t p = (ad * bd) + signextshift(ad * (b & RMASK)) + signextshift(bd * (a & RMASK));
	return p;
}

//overflow problems?
//static inline sbc_fixed_t fxdiv(sbc_fixed_t a, sbc_fixed_t b) { return (a<<FRAC)/b; }
#endif

int sbc_init(sbc_t *sbc, unsigned long flags);
int sbc_decode(sbc_t *sbc, void *data, int count);
int sbc_encode(sbc_t *sbc, void *data, int count);
void sbc_finish(sbc_t *sbc);

#ifdef __cplusplus
}
#endif

#endif /* __SBC_H */
